"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var react_1 = require("react");
var MqlContext_1 = __importDefault(require("../context/MqlContext/MqlContext"));
var CreateMqlQuery_1 = __importDefault(require("../mutations/mql/CreateMqlQuery"));
var FetchMqlQueryTimeSeries_1 = __importDefault(require("../queries/mql/FetchMqlQueryTimeSeries"));
var MqlQueryTypes_1 = require("../queries/mql/MqlQueryTypes");
// This is the delay between the _response_ from the last query and the _start_ of the new query
var QUERY_POLLING_MS = 200;
// Time in milliseconds to wait for a query to respond successfully before showing a slow loading message to the user.
var LONG_FETCH_QUERY_ATTEMPT_MAX = 20000; // 20 seconds
// As we keep track of prevoius queries, we don't want the list to grow unbounded in long running browser sessions
var CANCELLED_QUERY_LIST_LENGTH = 10;
var EMPTY_CONSTRAINT = { constraint: null };
// This filters out invalid WHERE clauses the consequence of janky form state;
function clearEmptyConstraints(where) {
    if (!where)
        return EMPTY_CONSTRAINT;
    if (where.And) {
        var newAnd = where.And.filter(function (val) { return (val.values || []).length > 0; });
        return newAnd.length > 0 ? __assign(__assign({}, where), { And: newAnd }) : EMPTY_CONSTRAINT;
    }
    if (where.Or) {
        var newOr = where.Or.filter(function (val) { return (val.values || []).length > 0; });
        return newOr.length > 0 ? __assign(__assign({}, where), { Or: newOr }) : EMPTY_CONSTRAINT;
    }
    return where;
}
function getErrorMessage(e) {
    if (e.message)
        return e.message;
    if (e.networkError)
        return e.networkError.message;
    if (e.graphQLErrors)
        return e.graphQLErrors[0].message;
    return "Invalid MQL Query error message";
}
var initialState = {
    queryId: null,
    queryStatus: MqlQueryTypes_1.MqlQueryStatus.Pending,
    data: null,
    cancelledQueries: [],
    fetchStartTime: null,
    isTakingForever: false,
};
/*
    At first glance, it is counter-intuitive that we should have a mix of the Reducer pattern
    and the Hook pattern. The explanation for this complexity is that we have several fetches
    that require coordination.
      1. We post a mutation to create a query
      2. We use the query ID to poll for results
      3. None of this can take place until an MQL Server URL is supplied
    This reducer abstracts all this complexity.
*/
function mqlQueryReducer(state, action) {
    var _a;
    switch (action.type) {
        case "postQueryStart": {
            // This condition is triggered when there is already a queryId being fetched.
            // In this case we want to add "cancel" that query by adding it to the cancelledQueries list
            if (state.queryId) {
                return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, cancelledQueries: __spreadArray([
                        state.queryId
                    ], state.cancelledQueries.slice(CANCELLED_QUERY_LIST_LENGTH - 1)), data: null, fetchStartTime: Date.now(), isTakingForever: false, errorMessage: undefined });
            }
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Running, fetchStartTime: Date.now(), isTakingForever: false, data: null, errorMessage: undefined });
        }
        case "postQueryFail": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Failed, errorMessage: action.errorMessage });
        }
        case "postQuerySuccess": {
            return __assign(__assign({}, state), { queryId: action.queryId, errorMessage: undefined });
        }
        case "fetchResultsStart": {
            return __assign({}, state);
        }
        case "fetchResultsRunning": {
            var now = Date.now();
            var diff = now - (state.fetchStartTime || 0);
            return __assign(__assign({}, state), { isTakingForever: diff >= LONG_FETCH_QUERY_ATTEMPT_MAX, errorMessage: undefined });
        }
        case "fetchResultsFail": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Failed, isTakingForever: false, errorMessage: action.errorMessage });
        }
        case "fetchResultsSuccess": {
            return __assign(__assign({}, state), { queryStatus: MqlQueryTypes_1.MqlQueryStatus.Successful, data: (_a = action.data) === null || _a === void 0 ? void 0 : _a.mqlQuery, fetchStartTime: null, isTakingForever: false, errorMessage: undefined });
        }
        default: {
            throw new Error();
        }
    }
}
// This custom hook consists of one useCallback and two useHooks that should asynchronously handle all scenarios for this chained
// data fetching we are doing. It should do it in a high performance way and in a way that is resilient to race conditions if the
// user updates their request while a query is in flight
function useMqlQuery(_a) {
    var queryInput = _a.queryInput, metricName = _a.metricName, limit = _a.limit, skip = _a.skip;
    var _b = react_1.useContext(MqlContext_1.default), mqlClient = _b.mqlClient, useMutation = _b.useMutation, handleCombinedError = _b.handleCombinedError, mqlServerUrl = _b.mqlServerUrl;
    var _c = react_1.useReducer(mqlQueryReducer, initialState), state = _c[0], dispatch = _c[1];
    var _d = useMutation(CreateMqlQuery_1.default), _e = _d[0], createMqlQuery = _d[1];
    react_1.useEffect(function () {
        if (!metricName || !mqlServerUrl || skip) {
            return;
        }
        var formState = {};
        if (queryInput) {
            formState = queryInput;
        }
        dispatch({ type: "postQueryStart" });
        createMqlQuery({
            metrics: [metricName],
            groupBy: formState.groupBy || [],
            where: clearEmptyConstraints(formState.where),
            addTimeSeries: true,
        }).then(function (_a) {
            var _b, _c;
            var data = _a.data, error = _a.error;
            if ((_b = data === null || data === void 0 ? void 0 : data.createMqlQuery) === null || _b === void 0 ? void 0 : _b.id) {
                dispatch({
                    type: "postQuerySuccess",
                    queryId: (_c = data === null || data === void 0 ? void 0 : data.createMqlQuery) === null || _c === void 0 ? void 0 : _c.id,
                });
            }
            else if (error) {
                dispatch({
                    type: "postQueryFail",
                    errorMessage: error && error.message,
                });
                handleCombinedError(error);
            }
        });
    }, [queryInput, metricName, mqlServerUrl, skip]);
    // Note: Extensive effort has gone into using useQuery here rather than an imperative fetch,
    // but we have not managed to avoid an infinite loop.
    var fetchResults = react_1.useCallback(function () {
        dispatch({ type: "fetchResultsStart" });
        return mqlClient
            .query(FetchMqlQueryTimeSeries_1.default, {
            queryId: state.queryId || "",
        })
            .toPromise()
            .then(function (_a) {
            var data = _a.data, error = _a.error;
            if (error) {
                dispatch({
                    type: "fetchResultsFail",
                    errorMessage: getErrorMessage(error),
                });
                handleCombinedError(error);
            }
            if (!(data === null || data === void 0 ? void 0 : data.mqlQuery)) {
                return;
            }
            var _b = data.mqlQuery, status = _b.status, id = _b.id, result = _b.result;
            if (id && state.cancelledQueries.includes(id)) {
                return;
            }
            if (status === MqlQueryTypes_1.MqlQueryStatus.Successful) {
                dispatch({
                    type: "fetchResultsSuccess",
                    data: data,
                    limit: limit,
                    handleCombinedError: handleCombinedError,
                });
            }
            if (status === MqlQueryTypes_1.MqlQueryStatus.Running) {
                window.setTimeout(function () {
                    dispatch({ type: "fetchResultsRunning" });
                    fetchResults();
                }, QUERY_POLLING_MS);
            }
        });
    }, [state.cancelledQueries, limit, state.queryId]);
    react_1.useEffect(function () {
        if (skip ||
            !state.queryId ||
            state.cancelledQueries.includes(state.queryId)) {
            return;
        }
        fetchResults();
    }, [skip, state.queryId, state.cancelledQueries]);
    return state;
}
exports.default = useMqlQuery;
